package fpp.compiler.codegen

import java.time.Year

/** Write a Cpp doc as hpp or cpp */
trait CppDocWriter extends CppDocVisitor with LineUtils {

  case class Input(
    /** The hpp file */
    hppFile: CppDoc.HppFile,
    /** The cpp file name */
    cppFileName: String,
    /** The list of enclosing class names, backwards. A class name may include :: */
    classNameList: List[String] = Nil,
  ) {

    /** Get the enclosing class name, including any qualifier */
    def getEnclosingClassQualified: String = classNameList.reverse.mkString("::")
 
    /** Get the enclosing class name with no qualifier */
    def getEnclosingClassUnqualified: String = classNameList.head.split("::").reverse.head

  }

  def default(in: Input) = Nil

  /** Convert a param to a string with no trailing comma */
  def paramString(p: CppDoc.Function.Param): String

  /** Convert a param to a string with a trailing comma */
  def paramStringComma(p: CppDoc.Function.Param): String

  /** Visit a CppDoc */
  def visitCppDoc(cppDoc: CppDoc): Output

  /** Write a parameter with no trailing comma as a line */
  final def writeParam(p: CppDoc.Function.Param): Line = line(paramString(p))

  /** Write a parameter with a trailing comma as a line */
  final def writeParamComma(p: CppDoc.Function.Param): Line = line(paramStringComma(p))

  type Output = List[Line]

}

object CppDocWriter extends LineUtils {

  /** Write a banner comment */
  def writeBannerComment(comment: String): List[Line] = {
    def banner =
      line("// ----------------------------------------------------------------------")
    (Line.blank :: banner :: writeCommentBody(comment)) :+ banner
  }

  /** Write a comment */
  def writeComment(comment: String): List[Line] = Line.blank :: writeCommentBody(comment)

  /** Write an optional Doxygen comment */
  def writeDoxygenCommentOpt(commentOpt: Option[String]): List[Line] = commentOpt match {
    case Some(comment) => writeDoxygenComment(comment)
    case None => Line.blank :: Nil
  }
    
  /** Write a Doxygen comment */
  def writeDoxygenComment(comment: String): List[Line] = 
    Line.blank ::lines(comment).map(Line.join(" ")(line("//!"))_)
    
  /** Write a comment body */
  def writeCommentBody(comment: String): List[Line] = lines(comment).map(Line.join(" ")(line("//"))_)

  /** Left align a compiler directive */
  def leftAlignDirective(line: Line): Line =
    if (line.string.startsWith("#")) Line(line.string) else line

  /** Write a header banner */
  def writeBanner(title: String, toolName: Option[String], description: String): List[Line] = lines(
    s"""|// ======================================================================
        |// \\title  $title
        |// \\author Generated by ${toolName.getOrElse("fpp tools")}
        |// \\brief  $description
        |// ======================================================================"""
  )

  /** Write a function body */
  def writeFunctionBody(body: List[Line]): List[Line] = {
    val bodyLines = body.length match {
      case 0 => Line.blank :: Nil
      case _ => body.map(indentIn(_))
    }
    line("{") :: (bodyLines :+ line("}"))
  }

}
